// Copyright (C) 2020 Intel Corporation

// SPDX-License-Identifier: MIT

#include <CL/sycl.hpp>
#include <chrono>
using namespace sycl;

// Appropriate values depend on your HW
constexpr int BLOCK_SIZE = 42;
constexpr int NUM_BLOCKS = 2500;
constexpr int N = NUM_BLOCKS * BLOCK_SIZE;
constexpr int TEST_LOOP = 5;
constexpr int GUESS_PAGE = 1024;
using ns = std::chrono::nanoseconds;

int main() {
  queue Q{gpu_selector{}};
  queue Q2{gpu_selector{}};
  int *data = malloc_shared<int>(N, Q);
  int *data2 = malloc_shared<int>(N, Q2);
  int *read_only_data = malloc_shared<int>(BLOCK_SIZE, Q);
  int *read_only_data2 = malloc_shared<int>(BLOCK_SIZE, Q2);

  // Never updated after initialization
  for (int i = 0; i < BLOCK_SIZE; i++){
    read_only_data[i] = i;
    read_only_data2[i] = i;
  }

  for (int i = 0; i < N; i++)
    data[i] = 0; // done on the host side
  for (int i = 0; i < N; i++)
    data2[i] = 0; // done on the host side
  // Mark this data as "read only" so the runtime can copy it
  // to the device instead of migrating it from the host.
  // Real values will be documented by your DPC++ backend.
  int HW_SPECIFIC_ADVICE_RO = 0;

 
  Q.mem_advise(read_only_data, BLOCK_SIZE,(pi_mem_advice)HW_SPECIFIC_ADVICE_RO);
  Q2.mem_advise(read_only_data2, BLOCK_SIZE,(pi_mem_advice)HW_SPECIFIC_ADVICE_RO);
  auto start = std::chrono::steady_clock::now();

   for (int b = 0; b <  N; b+=GUESS_PAGE) {
    Q.single_task([=](){
      // only need to read first data of page
      // to get the page into device memory
                  volatile int x =data[b]; });  
    }
  Q.wait();

  auto duration = std::chrono::steady_clock::now() - start;
  auto time = std::chrono::duration_cast<ns>(duration).count();
  std::cout << "setup-time  :" << time/1.0E9 << "\n";

  
  start = std::chrono::steady_clock::now();

   for (int b = 0; b <  N; b+=GUESS_PAGE) {
    Q2.prefetch(&data2[b], BLOCK_SIZE); // prefetch hint

    }
    Q2.wait();
   auto duration2 = std::chrono::steady_clock::now() - start;
   time = std::chrono::duration_cast<ns>(duration2).count();
  std::cout << "pfsetup-time:" << time/1.0E9 << "\n";

  //  give prefetch time to work ... delay same time as used by reads above
  start = std::chrono::steady_clock::now();

  do{
    duration2 = std::chrono::steady_clock::now() - start;
  } while (duration2 < duration);

  for (int j=0;j<TEST_LOOP;j++){
  auto start = std::chrono::steady_clock::now();

  for (int b = 0; b < NUM_BLOCKS; b++) {
    Q.parallel_for(range{BLOCK_SIZE},//e,
                   [=](id<1> i) { data[b * BLOCK_SIZE + i] += read_only_data[i]; });

  }
  Q.wait();
  auto duration = std::chrono::steady_clock::now() - start;
  auto time = std::chrono::duration_cast<ns>(duration).count();
  std::cout << "time  :" << time/1.0E9 << "\n";

    Q2.wait();
  start = std::chrono::steady_clock::now();
  for (int b = 0; b < NUM_BLOCKS; b++) {
    Q2.parallel_for(range{BLOCK_SIZE},//e,
                   [=](id<1> i) { data2[b * BLOCK_SIZE + i] += read_only_data2[i]; });

  }
  Q2.wait();

  duration = std::chrono::steady_clock::now() - start;
  time = std::chrono::duration_cast<ns>(duration).count();
  std::cout << "pftime:" << time/1.0E9 << "\n";

    }

  bool passed = true;
  for (int i=0;i<N;i++){
    int x = data[i];
    int ii = i%BLOCK_SIZE;
    int j= x/TEST_LOOP;
    passed = passed && j==ii;
  }

    for (int i=0;i<N;i++){
    int x = data2[i];
    int ii = i%BLOCK_SIZE;
    int j= x/TEST_LOOP;
    passed = passed && j==ii;
    }

  std::cout << (passed? "pass" : "fail") <<"\n";

  free(data, Q);
  free(read_only_data, Q);
  free(data2, Q2);
  free(read_only_data2, Q2);
  return 0;
}
